function [fisherMat,paramOrder] = getFisher3D...
    (t1,t2,t3,ampVec,fVec1,fVec2,fVec3,...
    betaVec1,betaVec2,betaVec3,phiVec...
    ,sigma2,parameterorder)

% Returns the Fisher Information matrix corresponding to the 3D sampling
% time (t1,t2,t3) for a signal consisting of K 3D complex damped sinusoids.
%
% INPUT
% t1            -       sampling time in first dimension (scalar value).
% t2            -       sampling time in second dimension (scalar value).
% ampVec        -       vector of magnitudes of size K-times-1.
% fVec1         -       vector of frequencies in the first dimension, size
%                       K-times-1.
% fVec2         -       vector of frequencies in the second dimension, size
%                       K-times-1.
% fVec2         -       vector of frequencies in the third dimension, size
%                       K-times-1.
% betaVec1      -       vector of dampings in the first dimension.
% betaVec2      -       vector of dampings in the second dimension.
% betaVec3      -       vector of dampings in the third dimension.
% phiVec        -       vector of initial phases.
% sigma2        -       noise variance.
% parameterorder -      string, either 'per_sinusoid' or 'by_parameter'.
%                       Determines how the FIM is structure. If
%                       'by_parameter', then each parameter type comes in a
%                       block: all amplitudes, all beta1, all beta2, all
%                       beta3, all f1, all f2, all f3, all phases. 
%                       If 'per_sinusoid', then the
%                       parameters of each sinusoid come in a collected
%                       block, in the order amp, beta1, beta2, beta3, f1, 
%                       f2, f3, phase. Default is 'by_parameter'.
%
% OUTPUT
% fisherMat     -       Fisher information matrix. Parameters are ordered
%                       either 'by_parameter' or 'by_sinusoid'
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% This version: 9 November 2021. 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin < 10
   parameterorder = 'by_parameter'; 
else
   if ~strcmp(parameterorder,'by_parameter') && ~strcmp(parameterorder,'by_sinusoid')
       error('Not valid choice of parameter order')
   end
end
K = length(ampVec);

yFunc = @(a,f1,f2,f3,beta1,beta2,beta3, phi,m1,m2,m3) [a*exp(1i*phi)*...
    exp(2i*pi*f1*m1-beta1*m1)*exp(2i*pi*f2*m2-beta2*m2)*exp(2i*pi*f3*m3-beta3*m3)];
% Per-component gradient, i.e., per sinusoid
% Parameter order: amplitude, beta1, beta2, freq1, freq2, phase.
gradH = @(a,f1,f2,f3,beta1,beta2,beta3,phi,m1,m2,m3,y) [y/a;...
    -y*m1;...
    -y*m2;...
    -y*m3;...
    2i*pi*m1*y;...
    2i*pi*m2*y;...
    2i*pi*m3*y;...
    1i*y];

fullMat = [];
for kSine = 1:K
    a = ampVec(kSine);
    f1 = fVec1(kSine);
    f2 = fVec2(kSine);
    f3 = fVec3(kSine);
    beta1 = betaVec1(kSine);
    beta2 = betaVec2(kSine);
    beta3 = betaVec3(kSine);
    phi = phiVec(kSine);

    
    y = yFunc(a,f1,f2,f3,beta1,beta2,beta3,phi,t1,t2,t3);
    gradHEval = gradH(a,f1,f2,f3,beta1,beta2,beta3,phi,t1,t2,t3,y);
    fullMat = [fullMat, gradHEval];
end

if strcmp(parameterorder,'by_parameter')
    fullVec = fullMat.';
    paramOrder = {'by_parameter','amp','damp1','damp2','damp3','freq1','freq2','freq3','phase'};
else
    paramOrder = {'by_sinusoid','amp','damp1','damp2','damp3','freq1','freq2','freq3','phase'};
end

fullVec = fullVec(:);
fisherMat = 2/sigma2*real(fullVec*fullVec');
end