function [w,wInd,terminationFlag,argsOut] = ...
    sampling_scheme_selector(Fcell,gamma,psiVec,dualGapTol)
%
% Primal-dual solver for the sample selection Menproblem described in
% "DESIGNING SAMPLING SCHEMES FOR MULTI-DIMENSIONAL DATA" by Johan Sw?rd, 
% Filip Elvander, and Andreas Jakobsson. The paper is published in Elsevier
% Signal Processing, 2018, Volume 150, pages 1-10.
% DOI: 10.1016/j.sigpro.2018.03.011
%
%
%
% INPUTS (REQUIRED)
% Fcell             -       cell of size N-times-1, where N is the number
%                           of candidate signal samples, containing the
%                           Fisher information matrices of the signal
%                           samples.
% gamma             -       the number of desired signal samples, i.e., a
%                           positive integer smaller than or equal to N.
%                           Corresponds to the \gamma in the above
%                           referenced paper.
%
% INPUTS (OPTIONAL)
% psiVec            -       weighting vector of size P-times-1, where P is
%                           the number of signal parameters (note that each
%                           Fisher information matrix should be of size
%                           P-times-P), determining the weighting of the
%                           signal parameters. This corresponds to the
%                           variables \psi in the above referenced paper.
%                           NOTE: the entries of psiVec must be
%                           non-negative for bounded problems. Leave as
%                           empty is all parameters should be weighted
%                           equally.
% dualGapTol        -       convergence tolerance. The duality gap is an
%                           upper bound for the potential improvement of
%                           the objective function, i.e., the estimation
%                           accuracy, as measured by the Cram?r-Rao lower
%                           bound, may not improve more than this.
%
% OUTPUTS
% w                 -       vector if size N-times-1 consisting of zeros
%                           and ones. A one indicates that the sample is
%                           selected, whereas a zero indicates that it is
%                           not selected. Corresponds to the w in above
%                           referenced paper. NOTE: this solver is
%                           implemented so that the entries of w always are
%                           either zero or one.
% wInd              -       the indices of w corresponding to entries equal
%                           to one.
% terminationFlag   -       a flag indicating the cause of algorithm
%                           termination:
%                               1 <-> the algorithm has terminated due to
%                                     convergence of the primal and dual
%                                     objective functions, as determined by
%                                     the tolerance on the duality gap.
%                               0 <-> the algorithm has terminated due to
%                                     no step being possible as to increase
%                                     the value of the dual objective
%                                     function.
%                              -1 <-> the algorithm has terminated due to
%                                     the outer iteration limit has been
%                                     reached.
%                             If "0" or "-1" is returned, you may check the
%                             potential improvement of the objective by
%                             considering the field "finalDualityGap" in
%                             the output "argsOut". If this is not
%                             satisfactorily small, you may want to change
%                             the default step size alpha0 below, as to
%                             improve the convergence of the algorithm, or
%                             increase maxIter, the number of allowed
%                             iterations.
% argsOut           -       struct containing the primal optimization
%                           variables w and mu (same as in the above
%                           referenced paper), the dual variabel,
%                           termination flags, the final duality gap, as
%                           well as the trajectories of the primal and dual
%                           objective functions and their corresponding
%                           duality gap. Also, a lower bound for the
%                           achievable primal objective function is one
%                           additonal sample is included, i.e., if gamma is
%                           increased by one, based on the final dual
%                           variable is included.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Implementation by: Johan Sw?rd and Filip Elvander, 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Start timer
tic

%%%%%%%%%%%%% SETTINGS THAT MAY BE CHANGED BY USER %%%%%%%%%%%%%%%%%%%
% Set original step-length for the dual ascent method. If the algorithm
% suffers convergence problems, try changing this value.
alpha0 = 100;

% Default tolerance for the duality gap. The duality gap is an upper bound
% for the potential improvement of the primal function, i.e., at any
% iteration of the algorithm, the duality gap provides a bound for how much
% estimation accuracy may improve if the algorithm may continue to iterate.
if nargin < 4 || isempty(dualGapTol)
    dualGapTol = 1e-5;
end

% Iteration limits: maxIter is the maximum number of algorithm iterations,
% i.e., the maximum number of allowed steps in the dual function.
maxIter = 600;

% maxInnerIter is the maximum number of attempts to increase the dual
% function in a given algorithm iteration.
maxInnerIter = 40;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%% DETERMINE SOME VALUES FROM THE INPUT %%%%%%%%%%%%%%%%%%%%
% Determine the number of available sample points
N = length(Fcell);

% Determine the number of signal parameters
nbrOfParams = size(Fcell{1},1);

% Set missing values to default values
if nargin < 3 || isempty(psiVec)
    % Set equal weighting for all signal parameters
    psiVec = ones(nbrOfParams,1);
else
   psiVec = psiVec(:);
   if any(psiVec<0)
      disp('Negative weights not allowed: problem unbounded from below')
      return
   end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%% SET INITIAL VALUES FOR OPTIMIZATION VARIABLES %%%%%%%%%%%%
% First primal variable: sample selection vector
w = ones(N,1)/N*gamma;

% Dual variables
Gcell = cell(nbrOfParams,1);
for kParam = 1:nbrOfParams
    tempScaling = 0;
    if psiVec(kParam)>0
        tempScaling = 1/psiVec(kParam);
    end
    Gcell{kParam} = .001*tempScaling*eye(nbrOfParams+1);
end
F = 0;
FcellVec = cell(N,1);
for kWeight = 1:N
    F = F + Fcell{kWeight}*w(kWeight);
    FcellVec{kWeight} = vec(Fcell{kWeight});
end

% Compute reference objective value if all samples are included
FAllSamples = 0;
for kWeight = 1:N
    FAllSamples = FAllSamples + Fcell{kWeight};
end
referencePrimal = psiVec'*diag(inv(FAllSamples));

% Second primal variable: corresponding to estimation accuracy for each
% signal parameter
mu = diag(inv(F));

% Pre-allocate dual gradients
GgradientCell = cell(nbrOfParams,1);
GcandidateCell = cell(nbrOfParams,1);

% Intial dual value
oldDualVal = -Inf;

% Record the trajectories of the dual and primal objective functions
qValHist = zeros(maxIter,1);
primalValHist = zeros(maxIter,1);
innerIterHist = zeros(maxIter,1);

% Intialize algorithm termination flags
convergenceFlag = 0;
noStepPossibleFlag = 0;
maxIterReachedFlag = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%% ALGORITHM START %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
strFormat = '%4d |% .4E | % .4E | % .4E |\n';
fprintf('-----------------------------------------------\n')
fprintf('\tDetermining sampling scheme\n')
fprintf('-----------------------------------------------\n')
fprintf('Number of signal parameters: %d\n',nbrOfParams)
fprintf('Number of available samples: %d\n',N)
fprintf('Number of desired samples: %d\n',gamma)
fprintf('Duality gap tolerance: %.2E\n',dualGapTol)
fprintf('Prim. obj. if all samples included: %.4E\n',referencePrimal)
fprintf('***********************************************\n')
fprintf('iter | dual-gap   |  prim-obj   |? dual-obj   |\n')
fprintf('***********************************************\n')
fprintf('-----------------------------------------------\n')
% Outer algorithm loop: solve optimization problem by gradient ascent in
% the dual function.
for kIter = 1:maxIter
    
    % Check convergence
    if kIter >3 && (primalValHist(kIter-1)-qValHist(kIter-1)<dualGapTol)
        convergenceFlag = 1;
        break
    end
    % Prepare computation of gradient
    F = 0;
    for kWeight = 1:N
        F = F + Fcell{kWeight}*w(kWeight);
    end
    
    % Set initial stepsize.
    iterSwitch = 20;
    if kIter<=iterSwitch
        alpha = alpha0;
    else
        alpha = alpha0/1.00005^(kIter-iterSwitch);
    end
    ep = zeros(nbrOfParams,1);
    
    % Inner algorithm iteration: find feasible dual point improving dual
    % objective function. If no step possible, the algorithm terminates.
    innerIter = 1;
    while innerIter <= maxInnerIter
        % Compute gradient and take step
        for kParam = 1:nbrOfParams
            ep(kParam) = 1;
            gradTemp = -[F,ep;ep',mu(kParam)];
            GgradientCell{kParam} = gradTemp;
            ep(kParam) = 0;
        end
        % Take step
        for kParam = 1:nbrOfParams
            GgradTemp = GgradientCell{kParam};
            % Take step and project on psd cone
            GCandTemp = Gcell{kParam} + alpha*GgradTemp;
            
            %%%%%%%
            GCandTemp = projectOnPsdCone_TEST(GCandTemp,psiVec(kParam));
            %%%%%%%%
            GcandidateCell{kParam} = GCandTemp;
        end
        
        % Compute corresponding dual value
        [qVal,wCand,uCand,infFlag] = evaluateDual(GcandidateCell,Fcell,FcellVec,...
            nbrOfParams,N,gamma,psiVec);
        
        if infFlag || qVal <= oldDualVal+1e-10
            % Step is not dual feasible or leads to decrease in function
            % value -> decrease step size.
            alpha = alpha/10;
            innerIter = innerIter+1;
        else
            % Step is feasible and increases function value -> accept step.
            Gcell = GcandidateCell;
            
            % Smooth primal variables for faster convergence.
            w = kIter*w/(kIter+1) + wCand/(kIter+1);
            mu = kIter*mu/(kIter+1) + uCand/(kIter+1);
                        
            % Save dual and primal objective values.
            oldDualVal = qVal;
            qValHist(kIter) = qVal;
            primalValHist(kIter) = uCand'*psiVec;
            
            break
        end 
    end
    innerIterHist(kIter) = innerIter;
    
    
    % If no step possible, terminate
    if innerIter>=maxInnerIter
        noStepPossibleFlag = 1;
        break
    end
    
    fprintf(strFormat,kIter,primalValHist(kIter)-qValHist(kIter),...
        primalValHist(kIter),qValHist(kIter));
end

% Determine lower bound of primal objective function if one additional
% sample is included by computing dual function in present point and
% increasing gamma by 1.
lowerBoundOneAdditonalSample = evaluateDual(GcandidateCell,Fcell,FcellVec,...
            nbrOfParams,N,gamma+1,psiVec);

if kIter == maxIter
    maxIterReachedFlag = 1;
    w = wCand;
    mu = uCand;
else
    w = wCand;
    mu = uCand;
    primalValHist = primalValHist(1:kIter-1);
    qValHist = qValHist(1:kIter-1);
    innerIterHist = innerIterHist(1:kIter-1);
end

% Termination flag:
% 1 <-> convergence 
if convergenceFlag
   terminationFlag = 1;
   terminationMessage = 'Dual-primal convergence.';
end

% 0 <-> no step possible 
if noStepPossibleFlag
   terminationFlag = 0;
   terminationMessage = 'No dual step possible.';
end

% -1 <-> max iteration reached
if maxIterReachedFlag
   terminationFlag = -1;
   terminationMessage = 'Maximum number of iterations reached.';
end

fprintf('-----------------------------------------------\n')
fprintf('Algorithm termination.\n')
fprintf('Termination flag: %d.\n',terminationFlag)
fprintf('-----------------------------------------------\n')
% Stop timer
tOut = toc;
%%%%%%%%%%%%%%%%%%%%%%%% ALGORITHM END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%% ASSIGN OUTPUT VALUES %%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fprintf('Cause of termination: %s\n',terminationMessage)
fprintf('Number of iterations: %d\n',kIter)
fprintf('Time taken: %.3f seconds\n',tOut)
fprintf('Final primal objective: %.4E\n',primalValHist(end))
fprintf('Final duality gap: %.4E\n',primalValHist(end)-qValHist(end))
fprintf('Lower primal bound for %d samples: %.4E\n',gamma+1,lowerBoundOneAdditonalSample)
fprintf('-----------------------------------------------\n')
% Find indices of selected samples
wInd = sort(find(w==1));

% Struct containing primal and dual variables, termination flags, as well 
% as the trajectories of the primal and dual objective functions
argsOut = struct;
argsOut.w = w;
argsOut.wInd = wInd;
argsOut.mu = mu;
argsOut.dualVariable = Gcell;
argsOut.terminationFlag = terminationFlag;
argsOut.convergenceFlag = convergenceFlag;
argsOut.noStepPossibleFlag = noStepPossibleFlag;
argsOut.maxIterReachedFlag = maxIterReachedFlag;
argsOut.finalDualityGap = primalValHist(end)-qValHist(end);
argsOut.primalObjectiveHist = primalValHist;
argsOut.dualObjectiveHist = qValHist;
argsOut.dualityGapHist = primalValHist-qValHist;
argsOut.lowerBoundOneAdditionalSample = lowerBoundOneAdditonalSample;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


end


%%%%%%%%%  AUXILLIARY FUNCTIONS - PROJECTION ON PSD CONE %%%%%%%%%%%%%%%%%
function Gproj = projectOnPsdCone_TEST(G,psi)
% Project on positive semi-definite cone
P = zeros(size(G));
P(end,end) = 1;
X = G;
const_tol = 1e-8;


%%%%%% with initialization
[S,D] = eig(X);
d_temp = diag(D);
nz_inds = find(d_temp>0);
U_test = S(:,nz_inds)*diag(d_temp(nz_inds))*S(:,nz_inds)';
const_deviance = U_test(end,end)-psi;
if const_deviance>0
    lambda_upper = 0;
    lambda_lower = -0.1;
    while const_deviance>0
        lambda_lower = lambda_lower*10;
        [S,D] = eig(X+lambda_lower*P);
        d_temp = diag(D);
        nz_inds = find(d_temp>0);
        U_test = S(:,nz_inds)*diag(d_temp(nz_inds))*S(:,nz_inds)';
        const_deviance = U_test(end,end)-psi;
    end
else
    lambda_lower = 0;
    lambda_upper = .1;
    while const_deviance<0
        lambda_upper = lambda_upper*10;
        [S,D] = eig(X+lambda_upper*P);
        d_temp = diag(D);
        nz_inds = find(d_temp>0);
        U_test = S(:,nz_inds)*diag(d_temp(nz_inds))*S(:,nz_inds)';
        const_deviance = U_test(end,end)-psi;
    end
end

%%%%%%%%%%%

iter = 0;
max_iter = 500;
const_deviance = 10;
while abs(const_deviance)>const_tol
    lambda_mid = 0.5*(lambda_lower+lambda_upper);
    
    [S,D] = eig(X+lambda_mid*P);
    d_temp = diag(D);
    nz_inds = find(d_temp>0);    
    U_test = S(:,nz_inds)*diag(d_temp(nz_inds))*S(:,nz_inds)';
    
    const_deviance = U_test(end,end)-psi;
    
    if const_deviance>0
       lambda_upper = lambda_mid; 
    else
        lambda_lower = lambda_mid;
    end
    iter = iter+1; 
    if iter>=max_iter
        disp('Projection routine did not converge') 
        break
    end
end

Gproj = U_test;


end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%  AUXILLIARY FUNCTIONS - EVALUATE DUAL FUNCTION %%%%%%%%%%%%%%%%%
function [qVal,w,mu,infFlag] = evaluateDual(Gcell,Fcell,FcellVec,...
    nbrOfParams,N,gamma,psiVec)
% Evaluates the dual function.

% Find weights w
etaVec = zeros(N,1);
G1tot = 0;
G2Vec = zeros(nbrOfParams,1);
gVec = zeros(nbrOfParams,1);
for kParam = 1:nbrOfParams
    Gp = Gcell{kParam};
    G1tot = G1tot + Gp(1:nbrOfParams,1:nbrOfParams);
    gVec(kParam) = Gp(end,end);
    G2Vec(kParam) = Gp(end,kParam);
    if gVec(kParam)>psiVec(kParam)+1e-7
        disp('Unbounded problem')
        infFlag = 1;
        qVal = -Inf;
        mu = [];
        w = [];
        return
    end
end
vecG1tot = vec(G1tot);
for kWeight = 1:N
    etaVec(kWeight) = sum(FcellVec{kWeight}.*vecG1tot); % Faster
end
[etaValsSorted,sortedI] = sort(etaVec,'descend');
nonPosInd = find(etaValsSorted<0,1,'first');
if isempty(nonPosInd)
    finalInd = gamma;
else
    finalInd = min(gamma,nonPosInd-1);
end
w = zeros(N,1);
w(sortedI(1:finalInd)) = 1;

% Compute inverse FIM and set mu
F = 0;
for kWeight = 1:finalInd
    F = F + w(sortedI(kWeight))*Fcell{sortedI(kWeight)};
end
invFIM = inv(F);
mu = diag(invFIM);

% Compute value of dual function
qVal = (psiVec-gVec)'*mu - w'*etaVec - 2*sum(G2Vec);
infFlag = 0;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



